package edu.cmu.cs.fluid.Tallyho.analysis;

import org.eclipse.jdt.core.*;
import org.eclipse.jdt.core.dom.*;

import edu.cmu.cs.fluid.Tallyho.analysis.util.JavaEditorMarker;
import edu.cmu.cs.fluid.Tallyho.analysis.util.SType;
import edu.cmu.cs.fluid.Tallyho.coe.*;

public final class EmptyCatchClause {
  /**
   * Log4j logger for this class
   */
  private static final org.apache.log4j.Logger LOG =
    org.apache.log4j.Logger.getLogger("TALLYHO.EmptyCatchClause");

  // Holds information about the current Java source file being examined
  private ICompilationUnit f_javaFile;
  private String f_javaOSFileName;
  private String f_javaFileSource;
  private CompilationUnit f_ast;

  // Tally information
  private int expCount = 0; // total number of catch blocks found
  private int badCount = 0; // no annotation and empty (same as results)
  private int rawCount = 0; // total count of empty exceptions (no heuristics)
  private int annCount = 0; // count that contain an annotation
  private int comCount = 0; // count that contain a comment, any comment
  private int annDoNothing = 0; // annotation "do nothing" count
  private int annIgnore = 0; // annotation "ignore" count
  private int annSwallow = 0; // annotation "swallow" count

  // Holds information about the search of the Catch block's AST search
  private int f_emptyStatementCount;
  private int f_emptyStatementSearchTotal;

  private class EmptyClauseVisitor extends ASTVisitor {
    public void preVisit(ASTNode node) {
      f_emptyStatementSearchTotal++; // keep a total count
    }
    public boolean visit(Block node) {
      f_emptyStatementCount++;
      return true;
    }
    public boolean visit(EmptyStatement node) {
      f_emptyStatementCount++;
      return true;
    }
    public boolean visit(NullLiteral node) {
      f_emptyStatementCount++;
      return true;
    }
    public boolean visit(Javadoc node) {
      f_emptyStatementCount++;
      return true;
    }
  }

  private class ExceptionVisitor extends ASTVisitor {
    public boolean visit(CatchClause node) {
      expCount++;
      boolean isEmpty = false;
      f_emptyStatementCount = 0;
      f_emptyStatementSearchTotal = 0;
      if (node.getBody().statements().size() == 0) {
        // found an empty exception handler
        isEmpty = true;
      } else {
        // check for nonsense like "; ; ;" or "; { ; ; }"
        node.getBody().accept(new EmptyClauseVisitor());
        if (f_emptyStatementCount == f_emptyStatementSearchTotal) {
          isEmpty = true;
        }
      }
      if (isEmpty) { // add our findings to search results
        rawCount++;
        // check if this empty exception block is annotated as being
        // OK (e.g., catch (Exception e) { //@ ignore }
        int sIndex = node.getBody().getStartPosition();
        int eIndex = sIndex + node.getBody().getLength();
        String blockCode =
          f_javaFileSource.substring(sIndex, eIndex).toLowerCase();
        boolean ignoreAnnotationFound = false;
        // -- STRUCTURED COMMENTS -- We are positive this case is OK
        // check for "//@ ignore"
        ignoreAnnotationFound
          |= blockCode.matches(
            "(.|\\n|\\r|\\f)*(//\\@|/\\*\\@)(\\s)*ignore(.|\\n|\\r|\\f)*");
        // check for "//@ swallow"
        ignoreAnnotationFound
          |= blockCode.matches(
            "(.|\\n|\\r|\\f)*(//\\@|/\\*\\@)(\\s)*swallow(.|\\n|\\r|\\f)*");
        //-- UNSTRUCTURED COMMENTS --  We are pretty sure this case is OK
        // check for "do nothing", "ignore", or "swallow"
        if (blockCode.indexOf("do nothing") != -1) {
          ignoreAnnotationFound = true;
          annDoNothing++;
        }
        if (blockCode.indexOf("ignore") != -1) {
          ignoreAnnotationFound = true;
          annIgnore++;
        }
        if (blockCode.indexOf("swallow") != -1) {
          ignoreAnnotationFound = true;
          annSwallow++;
        }
        if (ignoreAnnotationFound) {
          annCount++;
        }
        boolean hasComment = false;
        hasComment |= blockCode.indexOf("//") != -1;
        hasComment |= blockCode.indexOf("/*") != -1;
        if (hasComment) {
          comCount++;
          if (!ignoreAnnotationFound) {
            LOG.info("comment not annotation : " + blockCode);
          }
        }

        if (!ignoreAnnotationFound) {
          badCount++;
          // try to determine the exception type being ignored
          String exceptionName = "(unknown)";
          ITypeBinding e = node.getException().getType().resolveBinding();
          if (e != null) {
            SType se = SType.getInstance(e);
            if (se != null) {
              exceptionName = se.getName();
            }
          }
          new CoEResultsCollecter(
            "ignored exception(s)",
            "Ignored exception " + exceptionName,
            new JavaEditorMarker(
              f_ast.lineNumber(node.getStartPosition()),
              f_javaFile),
            exceptionName,
            null);
        }
      }
      return false; // don't bother visiting child nodes
    }
  }

  public boolean find(ICompilationUnit file, CompilationUnit ast)
    throws JavaModelException {
    this.f_ast = ast;
    f_javaFileSource = file.getSource();
    f_javaFile = file;
    f_javaOSFileName =
      f_javaFile.getResource().getLocation().makeAbsolute().toOSString();

    ast.accept(new ExceptionVisitor()); // do the actual search

    ast = null; // avoid memory leak
    f_javaFile = null;
    return false; // everything went OK
  }

  public void recordResults() {
    double perc;
    LOG.info("------------------");
    LOG.info("IGNORED EXCEPTIONS");
    LOG.info("------------------");
    LOG.info(" catch block uses(u) = " + expCount);
    LOG.info(" ignored exceptions(t) = " + rawCount);
    perc = (double) rawCount / (double) expCount;
    LOG.info(" - %u = " + perc);
    LOG.info(" ignored, contain a comment = " + comCount);
    perc = (double) comCount / (double) rawCount;
    LOG.info(" - %t = " + perc);
    LOG.info(" ignored, don't contain a comment = " + (rawCount - comCount));
    perc = (double) (rawCount - comCount) / (double) rawCount;
    LOG.info(" - %t = " + perc);
    LOG.info(" ignored, annotated = " + annCount);
    perc = (double) annCount / (double) rawCount;
    LOG.info(" - %t = " + perc);
    LOG.info(" ignored, unannated = " + badCount);
    perc = (double) badCount / (double) rawCount;
    LOG.info(" - %t = " + perc);
    LOG.info(" \"do nothing\" count = " + annDoNothing);
    LOG.info(" \"ignore\" count = " + annIgnore);
    LOG.info(" \"swallow\" count = " + annSwallow);
    expCount = 0;
    badCount = 0;
    rawCount = 0;
    annCount = 0;
    comCount = 0;
    annDoNothing = 0;
    annIgnore = 0;
    annSwallow = 0;
  }
}