package edu.cmu.cs.fluid.Tallyho.analysis;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;

import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.*;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jface.operation.IRunnableWithProgress;

import edu.cmu.cs.fluid.Tallyho.coe.CoEResultsCollecter;

public final class Majordomo implements IRunnableWithProgress {
  /**
   * Log4j logger for this class
   */
  private static final org.apache.log4j.Logger LOG =
    org.apache.log4j.Logger.getLogger("TALLYHO.Majordomo");

  private String msg = "Searching Java code for quality issues...";
  private IJavaModel javaModel; // The root of all Java information!
  private IJavaProject[] javaProjects;
  private IPackageFragmentRoot[][] packageFragmentRoots;
  private IPackageFragment[][] packageFragments;
  private ICompilationUnit[] javaCompilationUnits; // all lumped together

  private boolean analysisStartup(IProgressMonitor monitor)
    throws JavaModelException {
    // start up our progress monitor
    monitor.beginTask(msg, 2);
    monitor.subTask("Preparing to perform analysis");

    IWorkspaceRoot myWorkspaceRoot = ResourcesPlugin.getWorkspace().getRoot();

    javaModel = JavaCore.create(myWorkspaceRoot);

    if (javaModel == null) {
      // no java information in workspace
      return true; // analysis startup failed
    }
    monitor.worked(1); // show user our progress

    // count up how many Java source files we have to look at
    javaProjects = javaModel.getJavaProjects();
    if (javaProjects.length == 0) {
      // no Java projects found
      return true; // analysis startup failed
    }

    packageFragmentRoots = new IPackageFragmentRoot[javaProjects.length][];

    packageFragments = new IPackageFragment[javaProjects.length][];

    List javaCompilationUnitList = new ArrayList(1000);

    for (int projectIndex = 0;
      projectIndex < javaProjects.length;
      projectIndex++) {

      packageFragmentRoots[projectIndex] =
        javaProjects[projectIndex].getPackageFragmentRoots();

      packageFragments[projectIndex] =
        javaProjects[projectIndex].getPackageFragments();

      for (int fragmentIndex = 0;
        fragmentIndex < packageFragments[projectIndex].length;
        fragmentIndex++) {
        // add the Java files in this package fragment to our list
        javaCompilationUnitList.addAll(
          Arrays.asList(
            packageFragments[projectIndex][fragmentIndex]
              .getCompilationUnits()));
      }
    }
    // transfer the list of Java files to a type-safe array for later use
    javaCompilationUnits =
      (ICompilationUnit[]) (javaCompilationUnitList
        .toArray(new ICompilationUnit[javaCompilationUnitList.size()]));
    javaCompilationUnitList = null; // get rid of our list
    monitor.worked(1); // show user our progress
    return false; // everything went OK
  }

  public void run(IProgressMonitor monitor)
    throws InvocationTargetException, InterruptedException {

    Calendar startTime = Calendar.getInstance();
    LOG.info(
      "START TIME = "
        + startTime.get(Calendar.DAY_OF_YEAR)
        + " "
        + startTime.get(Calendar.HOUR_OF_DAY)
        + ":"
        + startTime.get(Calendar.MINUTE)
        + ":"
        + startTime.get(Calendar.SECOND));

    try {
      // ----------------
      // ANALYSIS STARTUP
      // ----------------
      if (analysisStartup(monitor)) { // it failed
        throw new InterruptedException(
          "unable to find or setup Java information "
            + " to analyze...does your project compile?");
      }
      if (monitor.isCanceled()) { // did the user press cancel?
        throw new InterruptedException("user cancel");
      }
      LOG.debug("analysis startup completed");

      // --------------------------------------
      // PASS ONE THROUGH ALL JAVA SOURCE FILES
      // --------------------------------------
      monitor.beginTask(msg, javaCompilationUnits.length);

      CoEResultsCollecter.reset(); // reset the CoE tree (reflects our results)

      EmptyCatchClause emptyCatchClause = new EmptyCatchClause();
//      ExtraImports extraImports = new ExtraImports();
      PoorVariableAbstraction poorVariableAbstraction =
        new PoorVariableAbstraction();

      for (int fileIndex = 0;
        fileIndex < javaCompilationUnits.length;
        fileIndex++) {
        try {
          // parse the java file
          CompilationUnit ast =
            AST.parseCompilationUnit(javaCompilationUnits[fileIndex], true);
          LOG.info(
            " ==== "
              + javaCompilationUnits[fileIndex].getElementName()
              + " ====");
          if (emptyCatchClause.find(javaCompilationUnits[fileIndex], ast)
//            || (extraImports.find(javaCompilationUnits[fileIndex], ast))
            || poorVariableAbstraction.find(
              javaCompilationUnits[fileIndex],
              ast)) {
            throw new InterruptedException(
              "trouble analyzing the java file "
                + javaCompilationUnits[fileIndex]
                  .getResource()
                  .getLocation()
                  .makeAbsolute()
                  .toOSString()
                + "...does this file compile?");
          }
          if (fileIndex % 100 == 0) {
            System.gc(); // clean up memory (or we WILL run out)
          }
          monitor.worked(1);
          // show user our progress
          // add up total number of quality issues noted so far
          int issueCount = CoEResultsCollecter.getIssueCount();
          monitor.subTask(
            (issueCount > 0
              ? "("
                + issueCount
                + " issue"
                + (issueCount > 1 ? "s found) " : " found) ")
              : "")
              + "Examined "
              + fileIndex
              + " of "
              + javaCompilationUnits.length
              + " files");
          if (monitor.isCanceled()) { // did the user press cancel?
            emptyCatchClause.recordResults();
//            extraImports.recordResults();
            poorVariableAbstraction.recordResults();
            throw new InterruptedException("user cancel");
          }
        } catch (Throwable e) {
          if (e instanceof InterruptedException) {
            throw (InterruptedException) e;
          }
          LOG.error("Exception caught...skipping part of file", e);
        }
      }
      emptyCatchClause.recordResults();
//      extraImports.recordResults();
      poorVariableAbstraction.recordResults();
      System.gc();
      monitor.done();

      Calendar finishTime = Calendar.getInstance();
      LOG.info(
        "FINISH TIME = "
          + finishTime.get(Calendar.DAY_OF_YEAR)
          + " "
          + finishTime.get(Calendar.HOUR_OF_DAY)
          + ":"
          + finishTime.get(Calendar.MINUTE)
          + ":"
          + finishTime.get(Calendar.SECOND));
      long duration =
        ((finishTime.get(Calendar.HOUR) * 60 * 60)
          + (finishTime.get(Calendar.MINUTE) * 60)
          + finishTime.get(Calendar.SECOND))
          - ((startTime.get(Calendar.HOUR) * 60 * 60)
            + (startTime.get(Calendar.MINUTE) * 60)
            + startTime.get(Calendar.SECOND));
      LOG.info("DURATION IN SEC = " + duration + " sec");
      LOG.info("DURATION IN MIN = " + duration / 60 + " min");

    } catch (Throwable e) {
      if (e instanceof InterruptedException) {
        throw (InterruptedException) e;
      }
      LOG.error("Global adaption failure (java.lang.Throwable)", e);
      throw new InvocationTargetException(e);
    }
  }
}