package edu.cmu.cs.fluid.Tallyho.analysis;

import java.util.*;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.*;

import edu.cmu.cs.fluid.Tallyho.analysis.util.*;
import edu.cmu.cs.fluid.Tallyho.coe.CoELink;
import edu.cmu.cs.fluid.Tallyho.coe.CoEResultsCollecter;

public final class PVADatabase {
  /**
   * Log4j logger for this class
   */
  private static final org.apache.log4j.Logger LOG =
    org.apache.log4j.Logger.getLogger("TALLYHO.PVADatabase");

  //-----------------------------------------------//
  // V A R I A B L E   L I S T   S T R U C T U R E //
  //-----------------------------------------------//

  private class FieldUse {
    SVariable field;
    List useLocations = new LinkedList(); // of JavaEditorMarker
  }

  private class MethodUse {
    SMethod method;
    List useLocations = new LinkedList(); // of JavaEditorMarker
  }

  private class AssignmentDep {
    SType variableType;
    List assignmentLocations = new LinkedList(); // of JavaEditorMarker
  }

  private class ActualParameterDep {
    SMethod method;
    SType formalParameterType;
    int parameterNumber;
    List callLocations = new LinkedList(); // of JavaEditorMarker
  }

  private class ReturnDep {
    SMethod method;
    SType returnType;
    List returnLocations = new LinkedList(); // of JavaEditorMarker
  }

  private class DeclaredVariable {
    boolean abstractionPossible = false;
    boolean localVariable = false;
    boolean privateVariable = false;
    boolean declarationFound = false;
    boolean isMethodActualParameter = false;
    ICompilationUnit javaFile = null; // file the variable is declared in
    SVariable variable = null;
    SType variableType = null;
    SMethod methodVariableIsActualParameterOf = null;
    int variableIsActualParameterNo = 0;
    JavaEditorMarker variableDeclarationLocation = null;
    // until declaration found
    List fieldUseSignature = new LinkedList(); // of FieldUse
    List methodUseSignature = new LinkedList(); // of MethodUse
    List assignmentSignature = new LinkedList(); // of AssignmentDep
    List actualParameterSignature = new LinkedList(); // of ActualParameterDep
    List returnSignature = new LinkedList(); // of ReturnDep
    Set compatibleInterfaces = new HashSet(); // of SType
    Set compatibleSuperclasses = new HashSet(); // of SType
  }

  /**
   * list of variables (local and global scope) under consideration
   * for using a more abstract interface
   */
  private Map f_variables = new HashMap(); // of DeclaredVariable

  //-------------------------------//
  // G E N E R A L   H E L P E R S //
  //-------------------------------//

  private void error(String errorMessage) {
    System.out.println("[PVADatabase] " + errorMessage);
  }

  private boolean isNull(Object o, String errorMessage) {
    if (o == null) {
      error(errorMessage);
      return true;
    }
    return false;
  }

  private boolean isInteresting(
    IVariableBinding variable,
    ITypeBinding variableType) {
    // check for bad parameters
    if (isNull(variable,
      "isInteresting() called with null parameter: variable")) {
      return false;
    }
    if (isNull(variableType,
      "isInteresting() called with null parameter: variableType")) {
      return false;
    }

    ITypeBinding t =
      (variableType.isArray() ? variableType.getElementType() : variableType);
    if (t == null) {
      return false;
    }
    SType st = SType.getInstance(t);
    if (st.isPrimitive()) {
      return false;
    }
    //    if (st.isSubtypeOf("java.lang.Throwable")) {
    //      return false;
    //    }
    //    if (st.isSubtypeOf("java.lang.String")) {
    //      return false;
    //    }
    return true;
  }

  private boolean isFromSource(IVariableBinding variable) {
    // check for bad parameters
    if (isNull(variable,
      "isFromSource() called with null parameter: variable")) {
      return false;
    }

    if (!variable.isField()) {
      return true; // it is a local variable so it must be from source code
    }
    ITypeBinding declaringClass = variable.getDeclaringClass();
    if (isNull(declaringClass,
      "can't get declaring class of variable "
        + DOMUtilities.toFullName(variable))) {
      // assume it is not from source, since we have no idea
      return false;
    } else {
      return declaringClass.isFromSource();
    }
  }

  /**
   * Searchs for a variable within the f_variables list and returns its
   * information if it is found, null is returned otherwise.
   */
  private DeclaredVariable lookupVariable(IVariableBinding variable) {
    // check for bad parameters
    if (isNull(variable,
      "lookupVariable() called with null parameter: variable")) {
      return null;
    }

    SVariable sv = SVariable.getInstance(variable);
    return (DeclaredVariable) f_variables.get(sv);
  }

  private DeclaredVariable lookupOrAddVariable(
    IVariableBinding variable,
    ITypeBinding variableType) {
    // check for bad parameters
    if (isNull(variable,
      "lookupOrAddVariable() called with null parameter: variable")) {
      return null;
    }
    if (isNull(variableType,
      "lookupOrAddVariable() called with null parameter: variableType")) {
      return null;
    }

    DeclaredVariable result = lookupVariable(variable);
    if (result != null) {
      return result; // we already know about this variable
    }
    // not found, so save in "f_variables"
    if (!isInteresting(variable, variableType)) {
      return null; // skip if not an "interesting" variable
    }
    // is the variable from source code?
    if (!isFromSource(variable)) {
      return null; // we are not interested in variables within JARS/etc.
    }
    result = new DeclaredVariable();
    result.variable = SVariable.getInstance(variable);
    result.variableType = SType.getInstance(variableType);
    result.localVariable = !variable.isField();
    result.privateVariable = (variable.getModifiers() & Modifier.PRIVATE) != 0;
    f_variables.put(result.variable, result);
    return result;
  }

  public void declareVariable(
    IVariableBinding variable,
    ITypeBinding variableType,
    JavaEditorMarker location) {
    varDecs++;
    // check for bad parameters
    if (isNull(variable,
      "declareVariable() called with null parameter: variable")) {
      return;
    }
    if (isNull(variableType,
      "declareVariable() called with null parameter: variableType")) {
      return;
    }
    if (isNull(location,
      "declareVariable() called with null parameter: location")) {
      return;
    }
    if (isInteresting(variable, variableType)) {
      // check if this variable exists in our list of variables
      DeclaredVariable v = lookupVariable(variable);
      if (v == null) {
        // not found, so we need to create it on our f_variables list
        v = new DeclaredVariable();
        v.variable = SVariable.getInstance(variable);
        v.variableType = SType.getInstance(variableType);
        v.localVariable = !variable.isField();
        v.privateVariable = (variable.getModifiers() & Modifier.PRIVATE) != 0;
        f_variables.put(v.variable, v);
      }
      v.variableDeclarationLocation = location;
      v.javaFile = location.getJavaFile();
      v.declarationFound = true;
    }
  }

  public void addActualParameterInformation(
    IVariableBinding variable,
    IMethodBinding methodVariableIsActualParameterOf,
    int parameterNumber) {
    DeclaredVariable v = lookupVariable(variable);
    if (v == null) {
      return; // ignore the parameter information
    }
    v.isMethodActualParameter = true;
    v.methodVariableIsActualParameterOf =
      SMethod.getInstance(methodVariableIsActualParameterOf);
    v.variableIsActualParameterNo = parameterNumber;
  }

  public void addToUseSignature(
    IVariableBinding variable,
    ITypeBinding variableType,
    IBinding member,
    JavaEditorMarker location) {
    // check for bad parameters
    if (isNull(variable,
      "addToUseSignature() called with null parameter: variable")) {
      return;
    }
    if (isNull(variableType,
      "addToUseSignature() called with null parameter: variableType")) {
      return;
    }
    if (isNull(member,
      "addToUseSignature() called with null parameter: member")) {
      return;
    }
    if (isNull(location,
      "addToUseSignature() called with null parameter: location")) {
      return;
    }

    DeclaredVariable dv = lookupOrAddVariable(variable, variableType);
    if (dv == null) {
      return; // we aren't interested in this variable
    }
    switch (member.getKind()) {
      case IBinding.VARIABLE :
        IVariableBinding vb = (IVariableBinding) member;
        // check that the declaring class is our variable of interest
        // this ensures that we aren't dealing with ".length" on an
        // array or something strange
        if (vb.getDeclaringClass() != null) {
          if (SType.getInstance(vb.getDeclaringClass())
            != (dv.variableType.isArray()
              ? dv.variableType.getElementType()
              : dv.variableType)) {
            return;
          }
        }
        SVariable sv = SVariable.getInstance(vb);
        //System.out.println(dv.variable + " [Field Use Found] " + sv);
        boolean fieldAdded = false;
        Iterator i = dv.fieldUseSignature.iterator();
        while (i.hasNext()) {
          FieldUse fu = (FieldUse) i.next();
          if (fu.field == sv) {
            fu.useLocations.add(location);
            fieldAdded = true;
          }
        }
        if (!fieldAdded) {
          FieldUse fu = new FieldUse();
          fu.field = sv;
          fu.useLocations.add(location);
          dv.fieldUseSignature.add(fu);
        }
        break;
      case IBinding.METHOD :
        IMethodBinding mb = (IMethodBinding) member;
        SMethod sm = SMethod.getInstance(mb);
        //System.out.println(dv.variable + " [Method Use Found] " + sm);
        boolean methodAdded = false;
        Iterator i1 = dv.methodUseSignature.iterator();
        while (i1.hasNext()) {
          MethodUse mu = (MethodUse) i1.next();
          if (mu.method == sm) {
            mu.useLocations.add(location);
            methodAdded = true;
          }
        }
        if (!methodAdded) {
          MethodUse mu = new MethodUse();
          mu.method = sm;
          mu.useLocations.add(location);
          dv.methodUseSignature.add(mu);
        }
        break;
      default :
        error(
          "unknown binding type for suspected field/method use \""
            + member.getName()
            + "\"");
    }
  }

  public void addToAssignmentSignature(
    ITypeBinding lhsType,
    IVariableBinding rhsVariable,
    ITypeBinding rhsType,
    JavaEditorMarker location) {
    // check for bad parameters
    if (isNull(lhsType,
      "addToAssignmentSignature() called with null parameter: lhsType")) {
      return;
    }
    if (isNull(rhsVariable,
      "addToAssignmentSignature() called with null parameter: rhsVariable")) {
      return;
    }
    if (isNull(rhsType,
      "addToAssignmentSignature() called with null parameter: rhsType")) {
      return;
    }
    if (isNull(location,
      "addToAssignmentSignature() called with null parameter: location")) {
      return;
    }

    DeclaredVariable dv = lookupOrAddVariable(rhsVariable, rhsType);
    if (dv == null) {
      return; // we aren't interested in this variable
    }
    // check that the actual is not primitive (array.length)
    if (lhsType.isPrimitive()) {
      return;
    }
    // get element type rather than the array type
    if (lhsType.isArray()) {
      lhsType = lhsType.getElementType();
    }
    SType lhsSType = SType.getInstance(lhsType);
    //System.out.println(dv.variable + " [Assignment Found] " + lhsSVariable);
    boolean assignmentAdded = false;
    Iterator i = dv.assignmentSignature.iterator();
    while (i.hasNext()) {
      AssignmentDep ad = (AssignmentDep) i.next();
      if (ad.variableType == lhsSType) {
        ad.assignmentLocations.add(location);
        assignmentAdded = true;
      }
    }
    if (!assignmentAdded) {
      AssignmentDep ad = new AssignmentDep();
      ad.variableType = lhsSType;
      ad.assignmentLocations.add(location);
      dv.assignmentSignature.add(ad);
    }
  }

  public void addToActualParameterSignature(
    IVariableBinding variable,
    ITypeBinding variableType,
    IMethodBinding method,
    int parameterNumber,
    JavaEditorMarker location) {
    // check for bad parameters
    if (isNull(variable,
      "addToActualParameterSignature() called with null parameter: variable")) {
      return;
    }
    if (isNull(variableType,
      "addToActualParameterSignature() called with null parameter:"
        + " variableType")) {
      return;
    }
    if (isNull(method,
      "addToActualParameterSignature() called with null parameter: method")) {
      return;
    }
    if (isNull(location,
      "addToActualParameterSignature() called with null parameter: location")) {
      return;
    }

    DeclaredVariable dv = lookupOrAddVariable(variable, variableType);
    SMethod smethod = SMethod.getInstance(method);
    if (dv == null) {
      return; // we aren't interested in this variable
    }
    ITypeBinding formalParameterType =
      method.getParameterTypes()[parameterNumber];
    // check that the actual is not primitive (array.length)
    if (formalParameterType.isPrimitive()) {
      return;
    }
    // get element type rather than the array type
    if (formalParameterType.isArray()) {
      formalParameterType = formalParameterType.getElementType();
    }
    /*
    System.out.println(
      dv.variable
        + " [Actual Param Use Found] "
        + SType.getInstance(formalParameterType));
    */
    boolean parameterAdded = false;
    Iterator i = dv.actualParameterSignature.iterator();
    while (i.hasNext()) {
      ActualParameterDep apd = (ActualParameterDep) i.next();
      if ((apd.method == smethod)
        && (apd.parameterNumber == parameterNumber)) {
        apd.callLocations.add(location);
        parameterAdded = true;
      }
    }
    if (!parameterAdded) {
      ActualParameterDep apd = new ActualParameterDep();
      apd.method = smethod;
      apd.parameterNumber = parameterNumber;
      apd.formalParameterType = SType.getInstance(formalParameterType);
      apd.callLocations.add(location);
      dv.actualParameterSignature.add(apd);
    }
  }

  public void addToReturnSignature(
    IVariableBinding variable,
    ITypeBinding variableType,
    IMethodBinding method,
    ITypeBinding returnType,
    JavaEditorMarker location) {
    // check for bad parameters
    if (isNull(variable,
      "addToReturnSignature() called with null parameter: variable")) {
      return;
    }
    if (isNull(variableType,
      "addToReturnSignature() called with null parameter: variableType")) {
      return;
    }
    if (isNull(method,
      "addToReturnSignature() called with null parameter: method")) {
      return;
    }
    if (isNull(returnType,
      "addToReturnSignature() called with null parameter: returnType")) {
      return;
    }
    if (isNull(location,
      "addToReturnSignature() called with null parameter: location")) {
      return;
    }

    DeclaredVariable dv = lookupOrAddVariable(variable, variableType);
    SMethod smethod = SMethod.getInstance(method);
    if (dv == null) {
      return; // we aren't interested in this variable
    }
    // check that the actual is not primitive (array.length)
    if (returnType.isPrimitive()) {
      return;
    }
    // get element type rather than the array type
    if (returnType.isArray()) {
      returnType = returnType.getElementType();
    }
    /*
    System.out.println(
      dv.variable + " [return Use Found] " + SType.getInstance(returnType));
    */
    boolean returnAdded = false;
    Iterator i = dv.returnSignature.iterator();
    while (i.hasNext()) {
      ReturnDep rd = (ReturnDep) i.next();
      if (rd.method == smethod) {
        rd.returnLocations.add(location);
        returnAdded = true;
      }
    }
    if (!returnAdded) {
      ReturnDep rd = new ReturnDep();
      rd.method = smethod;
      rd.returnType = SType.getInstance(returnType);
      rd.returnLocations.add(location);
      dv.returnSignature.add(rd);
    }
  }

  public void findLocalPrivateVariableCompatible(boolean processGlobal) {
    Iterator i = f_variables.entrySet().iterator();
    while (i.hasNext()) {
      Map.Entry e = (Map.Entry) i.next();
      DeclaredVariable v = (DeclaredVariable) e.getValue();
      if (v.localVariable || v.privateVariable || processGlobal) {
        if (!v.declarationFound) {
          // strange that we didn't find the declaration for a local variable
          error(
            "never found declaration for the variable \"" + v.variable + "\"");
          i.remove();
        } else {
          // first check if this variable is a actual parameter declaration
          // in an overridden method (or part of a interface implementation)
          // if so we really can't safely change the type
          if (!isMethodOverideActualParameter(v)) {
            findCompatibleParentTypes(v, v.variableType);

            checkCompatibleWithAssignments(
              v.compatibleInterfaces,
              v.assignmentSignature);
            checkCompatibleWithAssignments(
              v.compatibleSuperclasses,
              v.assignmentSignature);

            checkCompatibleWithCalls(
              v.compatibleInterfaces,
              v.actualParameterSignature);
            checkCompatibleWithCalls(
              v.compatibleSuperclasses,
              v.actualParameterSignature);

            checkCompatibleWithReturns(
              v.compatibleInterfaces,
              v.returnSignature);
            checkCompatibleWithReturns(
              v.compatibleSuperclasses,
              v.returnSignature);
          }
          filterResults(v);
          v.abstractionPossible =
            (v.compatibleInterfaces.size() > 0)
              || (v.compatibleSuperclasses.size() > 0);
          if (v.abstractionPossible) {
            processResults(v); // put this out to results
          }
          i.remove(); // remove all local/private from active consideration
        }
      }
    }
  }

  private void filterResults(DeclaredVariable v) {
    //    for (Iterator i = v.compatibleSuperclasses.iterator(); i.hasNext();) {
    //      SType t = (SType) i.next();
    //      if (t.getName().equals("java.lang.Object")) {
    //        i.remove();
    //      }
    //    }
  }

  private boolean isMethodOverideActualParameter(DeclaredVariable v) {
    if (!v.isMethodActualParameter) {
      return false;
    } else {
      if (v.methodVariableIsActualParameterOf == null) {
        LOG.error(
          "null valuse for DeclaredVariable.methodVariableIsActualParameterOf "
            + "in isMethodOverideActualParameter()");
        return false;
      }
      SType t = v.methodVariableIsActualParameterOf.getDeclaringClass();
      if (t == null) {
        LOG.error(
          "null value for SMethod.getDeclaringClass() "
            + "in isMethodOverideActualParameter()");
        return false;
      }
      return t.containsMethod(v.methodVariableIsActualParameterOf);
    }
  }

  private void findCompatibleParentTypes(DeclaredVariable v, SType t) {
    SType[] interfaces = t.getInterfaces();
    for (int i = 0; i < interfaces.length; i++) {
      if (isCompatible(v, interfaces[i])) {
        v.compatibleInterfaces.add(interfaces[i]);
      }
      findCompatibleParentTypes(v, interfaces[i]);
    }
    if (t.getSuperclass() != null) {
      if (isCompatible(v, t.getSuperclass())) {
        v.compatibleSuperclasses.add(t.getSuperclass());
      }
      findCompatibleParentTypes(v, t.getSuperclass());
    }
  }

  private boolean isCompatible(DeclaredVariable v, SType t) {
    // check methods
    Iterator i = v.methodUseSignature.iterator();
    while (i.hasNext()) {
      SMethod m = ((MethodUse) i.next()).method;
      if (!t.containsMethod(m)) {
        return false;
      }
    }
    // check fields (only if not an interface)
    if (!v.variableType.isInterface()) {
      Set tFields = new HashSet(Arrays.asList(t.getDeclaredFields()));
      Iterator i1 = v.fieldUseSignature.iterator();
      while (i1.hasNext()) {
        SVariable f = ((FieldUse) i1.next()).field;
        if (!tFields.contains(f)) {
          return false;
        }
      }
    }
    return true;
  }

  private void checkCompatibleWithAssignments(
    Set types,
    List assignmentSignature) {
    Iterator it = types.iterator();
    while (it.hasNext()) {
      SType t = (SType) it.next();
      boolean remove = false;
      Iterator ia = assignmentSignature.iterator();
      while (ia.hasNext()) {
        AssignmentDep a = (AssignmentDep) ia.next();
        if (!t.isSubtypeOf(a.variableType)) {
          remove = true;
        }
      }
      if (remove) {
        it.remove();
      }
    }
  }

  private void checkCompatibleWithCalls(
    Set types,
    List actualParameterSignature) {
    Iterator it = types.iterator();
    while (it.hasNext()) {
      SType t = (SType) it.next();
      boolean remove = false;
      Iterator ia = actualParameterSignature.iterator();
      while (ia.hasNext()) {
        ActualParameterDep a = (ActualParameterDep) ia.next();
        if (!t.isSubtypeOf(a.formalParameterType)) {
          remove = true;
        }
      }
      if (remove) {
        it.remove();
      }
    }
  }

  private void checkCompatibleWithReturns(Set types, List returnSignature) {
    Iterator it = types.iterator();
    while (it.hasNext()) {
      SType t = (SType) it.next();
      boolean remove = false;
      Iterator ia = returnSignature.iterator();
      while (ia.hasNext()) {
        ReturnDep a = (ReturnDep) ia.next();
        if (!t.isSubtypeOf(a.returnType)) {
          remove = true;
        }
      }
      if (remove) {
        it.remove();
      }
    }
  }

  private long varDecs = 0;
  private int throwableCount = 0;
  private int stringCount = 0;
  private int onlyObject = 0;
  private int reportedObjects = 0;
  private int interfaceAllowed = 0;
  private int decConstInterfacesRemoved = 0;
  private Set decConstInterfaces = new HashSet();

  public void showCounters() {
    double perc;
    LOG.info("-------------------------");
    LOG.info("POOR VARIABLE ABSTRACTION");
    LOG.info("-------------------------");
    LOG.info(" variable dec usage(u) = " + varDecs);
    int total = throwableCount + stringCount + onlyObject + reportedObjects;
    LOG.info(" total(t) = " + total);
    perc = (double) total / (double) varDecs;
    LOG.info(" - %t = " + perc);
    LOG.info(" Throwable subclass = " + throwableCount);
    perc = (double) throwableCount / (double) total;
    LOG.info(" - %u = " + perc);
    LOG.info(" String instance = " + stringCount);
    perc = (double) stringCount / (double) total;
    LOG.info(" - %u = " + perc);
    LOG.info(" only Object allowed = " + onlyObject);
    perc = (double) onlyObject / (double) total;
    LOG.info(" - %u = " + perc);
    LOG.info(" reported as allowed = " + reportedObjects);
    perc = (double) reportedObjects / (double) total;
    LOG.info(" - %u = " + perc);
    LOG.info(" interface allowed = " + interfaceAllowed);
    LOG.info(" constent interfaces removed = " + decConstInterfacesRemoved);
    for (Iterator i = decConstInterfaces.iterator(); i.hasNext();) {
      LOG.info(" - interface removed = " + i.next());
    }
    varDecs = 0;
    throwableCount = 0;
    stringCount = 0;
    onlyObject = 0;
    reportedObjects = 0;
    interfaceAllowed = 0;
    decConstInterfacesRemoved = 0;
    decConstInterfaces = new HashSet();
  }

  private void processResults(DeclaredVariable v) {
    if (!v.declarationFound)
      return; // impossible, but just a double check
    boolean compInt = (v.compatibleInterfaces.size() > 0);
    if (compInt) {
      // pre-process to remove all interfaces with no methods
      // these are usually simply decorator interfaces or
      // constant interfaces and are always bad choices
      for (Iterator i = v.compatibleInterfaces.iterator(); i.hasNext();) {
        SType t = (SType) i.next();
        if (t.getDeclaredMethods().length == 0) {
          decConstInterfaces.add(t.getName());
          i.remove();
          decConstInterfacesRemoved++;
        }
      }
    }
    // check if we still have any interfaces left
    compInt = (v.compatibleInterfaces.size() > 0);
    boolean compSup = (v.compatibleSuperclasses.size() > 0);
    if (compInt || compSup) {
      if (v.variableType.isSubtypeOf("java.lang.Throwable")) {
        throwableCount++;
        return; // don't show this in the results just count it
      }
      if (v.variableType.isSubtypeOf("java.lang.String")) {
        stringCount++;
        return; // don't show this in the results just count it
      }
      CoELink root = new CoELink("root");
      // interfaces
      Iterator ii = v.compatibleInterfaces.iterator();
      while (ii.hasNext()) {
        SType t = (SType) ii.next();
        CoELink link = new CoELink(t.getName() + " is a compatible interface");
        link.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        root.addLink(link);
      }
      // superclasses
      Iterator is = v.compatibleSuperclasses.iterator();
      while (is.hasNext()) {
        SType t = (SType) is.next();
        CoELink link = new CoELink(t.getName() + " is a compatible superclass");
        link.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        root.addLink(link);
        // check if java.lang.Object is only possible replacement
        if (!compInt
          && v.compatibleSuperclasses.size() == 1
          && t.getName().equals("java.lang.Object")) {
          onlyObject++;
          return; // don't show this in the results just count it
        }
      }
      root.addLink(createSupportingInformation(v));
      new CoEResultsCollecter(
        "declared variable(s) with poor abstraction",
        v.variableType.getName()
          + " "
          + v.variable.getName()
          + " has "
          + (compInt
            ? v.compatibleInterfaces.size()
              + (v.compatibleInterfaces.size() > 1
                ? " compatible interfaces "
                : " compatible interface ")
              + (compSup ? "and " : "")
            : "")
          + (compSup
            ? v.compatibleSuperclasses.size()
              + (v.compatibleSuperclasses.size() > 1
                ? " compatible superclasses"
                : " compatible superclass")
            : ""),
        v.variableDeclarationLocation,
        v.variableType.getName(),
        root);
      reportedObjects++;
      if (compInt)
        interfaceAllowed++;
    }
  }

  private CoELink createSupportingInformation(DeclaredVariable v) {
    CoELink result = new CoELink("Supporting information:");
    result.setImage(CoEResultsCollecter.JAVA_PKG_IMG);
    // field use
    if (v.fieldUseSignature.size() > 0) {
      CoELink fieldUse = new CoELink(v.variable.getName() + "'s field use:");
      fieldUse.setImage(CoEResultsCollecter.JAVA_PROJ_IMG);
      result.addLink(fieldUse);
      for (Iterator i = v.fieldUseSignature.iterator(); i.hasNext();) {
        FieldUse f = (FieldUse) i.next();
        int useCount = f.useLocations.size();
        CoELink field =
          new CoELink(
            f.field.getName()
              + " referenced "
              + useCount
              + (useCount > 1 ? " times" : " time"));
        field.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        fieldUse.addLink(field);
        for (Iterator j = f.useLocations.iterator(); j.hasNext();) {
          JavaEditorMarker jm = (JavaEditorMarker) j.next();
          CoELink location =
            new CoELink(
              "line "
                + jm.getLineNumber()
                + ": "
                + DOMUtilities.toOSFileName(jm.getJavaFile()));
          location.setJavaEditorLocation(jm);
          location.setImage(CoEResultsCollecter.JAVA_COMP_UNIT_IMG);
          field.addLink(location);
        }
      }
    }
    // method use
    if (v.methodUseSignature.size() > 0) {
      CoELink methodUse = new CoELink(v.variable.getName() + "'s method use:");
      methodUse.setImage(CoEResultsCollecter.JAVA_PROJ_IMG);
      result.addLink(methodUse);
      for (Iterator i = v.methodUseSignature.iterator(); i.hasNext();) {
        MethodUse m = (MethodUse) i.next();
        int useCount = m.useLocations.size();
        CoELink method =
          new CoELink(
            m.method.getName()
              + "() invoked "
              + useCount
              + (useCount > 1 ? " times" : " time"));
        method.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        methodUse.addLink(method);
        for (Iterator j = m.useLocations.iterator(); j.hasNext();) {
          JavaEditorMarker jm = (JavaEditorMarker) j.next();
          CoELink location =
            new CoELink(
              "line "
                + jm.getLineNumber()
                + ": "
                + DOMUtilities.toOSFileName(jm.getJavaFile()));
          location.setJavaEditorLocation(jm);
          location.setImage(CoEResultsCollecter.JAVA_COMP_UNIT_IMG);
          method.addLink(location);
        }
      }
    }
    // assignment dependency
    if (v.assignmentSignature.size() > 0) {
      CoELink assignmentDep =
        new CoELink(
          v.variable.getName() + "'s assignment to another variable:");
      assignmentDep.setImage(CoEResultsCollecter.JAVA_PROJ_IMG);
      result.addLink(assignmentDep);
      for (Iterator i = v.assignmentSignature.iterator(); i.hasNext();) {
        AssignmentDep a = (AssignmentDep) i.next();
        int useCount = a.assignmentLocations.size();
        CoELink assignment =
          new CoELink(
            "Assigned to "
              + a.variableType.getName()
              + " "
              + useCount
              + (useCount > 1 ? " times" : " time"));
        assignment.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        assignmentDep.addLink(assignment);
        for (Iterator j = a.assignmentLocations.iterator(); j.hasNext();) {
          JavaEditorMarker jm = (JavaEditorMarker) j.next();
          CoELink location =
            new CoELink(
              "line "
                + jm.getLineNumber()
                + ": "
                + DOMUtilities.toOSFileName(jm.getJavaFile()));
          location.setJavaEditorLocation(jm);
          location.setImage(CoEResultsCollecter.JAVA_COMP_UNIT_IMG);
          assignment.addLink(location);
        }
      }
    }
    // actual parameter dependency
    if (v.actualParameterSignature.size() > 0) {
      CoELink aparmDep =
        new CoELink(
          v.variable.getName() + "'s use as an actual parameter to a method:");
      aparmDep.setImage(CoEResultsCollecter.JAVA_PROJ_IMG);
      result.addLink(aparmDep);
      for (Iterator i = v.actualParameterSignature.iterator(); i.hasNext();) {
        ActualParameterDep a = (ActualParameterDep) i.next();
        int useCount = a.callLocations.size();
        CoELink aparm =
          new CoELink(
            "Actual parameter "
              + a.parameterNumber
              + " in "
              + useCount
              + (useCount > 1 ? " calls to " : " call to ")
              + a.method.getName()
              + "() requiring formal parameter type "
              + a.formalParameterType.getName());
        aparm.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        aparmDep.addLink(aparm);
        for (Iterator j = a.callLocations.iterator(); j.hasNext();) {
          JavaEditorMarker jm = (JavaEditorMarker) j.next();
          CoELink location =
            new CoELink(
              "line "
                + jm.getLineNumber()
                + ": "
                + DOMUtilities.toOSFileName(jm.getJavaFile()));
          location.setJavaEditorLocation(jm);
          location.setImage(CoEResultsCollecter.JAVA_COMP_UNIT_IMG);
          aparm.addLink(location);
        }
      }
    }
    // return dependency
    if (v.returnSignature.size() > 0) {
      CoELink assignmentDep =
        new CoELink(v.variable.getName() + "'s use in a return statement:");
      assignmentDep.setImage(CoEResultsCollecter.JAVA_PROJ_IMG);
      result.addLink(assignmentDep);
      for (Iterator i = v.returnSignature.iterator(); i.hasNext();) {
        ReturnDep a = (ReturnDep) i.next();
        int useCount = a.returnLocations.size();
        CoELink retv =
          new CoELink(
            "return value of "
              + a.method.getName()
              + "() with return type of "
              + a.returnType.getName()
              + " "
              + useCount
              + (useCount > 1 ? " times" : " time"));
        retv.setImage(CoEResultsCollecter.CHAIN_LINK_IMG);
        assignmentDep.addLink(retv);
        for (Iterator j = a.returnLocations.iterator(); j.hasNext();) {
          JavaEditorMarker jm = (JavaEditorMarker) j.next();
          CoELink location =
            new CoELink(
              "line "
                + jm.getLineNumber()
                + ": "
                + DOMUtilities.toOSFileName(jm.getJavaFile()));
          location.setJavaEditorLocation(jm);
          location.setImage(CoEResultsCollecter.JAVA_COMP_UNIT_IMG);
          retv.addLink(location);
        }
      }
    }
    return result;
  }
}