package edu.cmu.cs.fluid.Tallyho.analysis;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.*;

import edu.cmu.cs.fluid.Tallyho.analysis.util.DOMUtilities;
import edu.cmu.cs.fluid.Tallyho.analysis.util.JavaEditorMarker;

/**
 * Runs the AST to uncover variable declarations and variable use
 * within a Java source file
 */
public final class PVAVariableDeclarationAndUse
  extends org.eclipse.jdt.core.dom.ASTVisitor {

  private CompilationUnit f_ast;
  private ICompilationUnit f_javaFile;
  private String f_javaOSFileName;
  private PVADatabase f_db;
  private IVariableBinding[] f_noVariableBindings = new IVariableBinding[0];

  public PVAVariableDeclarationAndUse(
    PVADatabase database,
    ICompilationUnit javaFile,
    CompilationUnit ast) {
    f_db = database;
    f_javaFile = javaFile;
    f_javaOSFileName = DOMUtilities.toOSFileName(javaFile);
    f_ast = ast;
  }

  private PVAVariableDeclarationAndUse() {
  }

  //-------------------------------//
  // G E N E R A L   H E L P E R S //
  //-------------------------------//

  private void error(String errorMessage) {
    System.out.println(
      "[PVAVariableDeclarationAndUse] "
        + errorMessage
        + " in file "
        + f_javaOSFileName);
  }

  private void log(String message) {
    System.out.println("[PVAVariableDeclarationAndUse] " + message);
  }

  private boolean isNull(Object o, String errorMessage) {
    if (o == null) {
      error(errorMessage);
      return true;
    }
    return false;
  }

  private JavaEditorMarker toJEM(ASTNode node) {
    return new JavaEditorMarker(
      f_ast.lineNumber(node.getStartPosition()),
      f_javaFile);
  }

  private IVariableBinding[] expressionToVariable(Expression expression) {
    // check for bad parameters
    if (isNull(expression,
      "expressionToVariable() called with null parameter: expression")) {
      return f_noVariableBindings;
    }
    Expression exp; // used for recursive calls
    switch (expression.getNodeType()) {
      case ASTNode.SIMPLE_NAME :
      case ASTNode.QUALIFIED_NAME :
        Name variableName = (Name) expression;
        IBinding b = variableName.resolveBinding();
        if (isNull(b,
          "can't resolve variable binding for NAME expression \""
            + expression
            + "\"")) {
          return f_noVariableBindings;
        }
        // OK try to get an actual name
        if (b.getKind() == IBinding.VARIABLE) {
          // it is a variable so look it up
          IVariableBinding[] result = {(IVariableBinding) b };
          return result;
        }
        break;
      case ASTNode.ARRAY_ACCESS :
        // ArrayAccess: *Expression* [ Expression ]
        exp = ((ArrayAccess) expression).getArray();
        if (isNull(exp,
          "no array found within ArrayAccess \"" + expression + "\"")) {
          return f_noVariableBindings;
        }
        return expressionToVariable(exp);
      case ASTNode.FIELD_ACCESS :
        // FieldAccess: Expression . *Identifier*
        exp = ((FieldAccess) expression).getName();
        if (isNull(exp,
          "no identifier found within FieldAccess \"" + expression + "\"")) {
          return f_noVariableBindings;
        }
        return expressionToVariable(exp);
      case ASTNode.PARENTHESIZED_EXPRESSION :
        // ParenthesizedExpression: ( *Expression* )
        exp = ((ParenthesizedExpression) expression).getExpression();
        if (isNull(exp,
          "no expression found within ParenthesizedExpression \""
            + expression
            + "\"")) {
          return f_noVariableBindings;
        }
        return expressionToVariable(exp);
      case ASTNode.CONDITIONAL_EXPRESSION :
        ConditionalExpression cd = (ConditionalExpression) expression;
        List result = new LinkedList();
        IVariableBinding[] thenPart =
          expressionToVariable(cd.getThenExpression());
        result.addAll(Arrays.asList(thenPart));
        IVariableBinding[] elsePart =
          expressionToVariable(cd.getElseExpression());
        result.addAll(Arrays.asList(elsePart));
        return (IVariableBinding[]) result.toArray(
          new IVariableBinding[result.size()]);
    }
    return f_noVariableBindings;
  }

  public boolean visit(ImportDeclaration node) {
    return false; // swallow so we don't get errors resolving libraries
  }

  public boolean visit(PackageDeclaration node) {
    return false; // swallow so we don't get errors resolving libraries
  }

  public boolean visit(TypeDeclaration node) {
    if (node.isInterface()) {
      return false; // skip interfaces don't want to see these
    } else {
      return true; // we care about classes
    }
  }

  //-------------------------------------------------------//
  // V A R I A B L E   D E C L E R A T I O N   S E A R C H //
  //-------------------------------------------------------//

  void saveVariable(VariableDeclarationFragment node) {
    if (isNull(node,
      "null VariableDeclarationFragment passed to saveVariable()")) {
      return;
    }
    IVariableBinding variable = node.resolveBinding();
    if (isNull(variable,
      "null variable in VariableDeclarationFragment \"" + node + "\"")) {
      return;
    }
    ITypeBinding type = variable.getType();
    if (isNull(type,
      "null type in VariableDeclarationFragment \"" + node + "\"")) {
      return;
    }
    // declare the variable (it may already have been encountered)
    f_db.declareVariable(variable, type, toJEM(node));

    // since we are interested in this variable check the initializer
    Expression initializer = node.getInitializer();
    if (initializer == null) {
      return; // no initializer
    }
    ITypeBinding rhsType = initializer.resolveTypeBinding();
    if (isNull(rhsType, "null rhs type in Initializer \"" + node + "\"")) {
      return;
    }
    IVariableBinding[] rhsVar = expressionToVariable(initializer);
    if (rhsVar.length == 0) {
      // some kind of expression we are not interested in
      return;
    }
    for (int i = 0; i < rhsVar.length; i++) {
      f_db.addToAssignmentSignature(type, rhsVar[i], rhsType, toJEM(node));
    }
  }

  public boolean visit(FieldDeclaration node) {
    // loop through all the declarations found
    Iterator i = node.fragments().iterator();
    while (i.hasNext()) {
      VariableDeclarationFragment vdf = (VariableDeclarationFragment) i.next();
      saveVariable(vdf); // save this delcaration
    }
    return true; // continue
  }

  public boolean visit(VariableDeclarationStatement node) {
    // loop through all the declarations found
    Iterator i = node.fragments().iterator();
    while (i.hasNext()) {
      VariableDeclarationFragment vdf = (VariableDeclarationFragment) i.next();
      saveVariable(vdf); // save this delcaration
    }
    return true; // continue
  }

  public boolean visit(VariableDeclarationExpression node) {
    // loop through all the declarations found
    Iterator i = node.fragments().iterator();
    while (i.hasNext()) {
      VariableDeclarationFragment vdf = (VariableDeclarationFragment) i.next();
      saveVariable(vdf); // save this delcaration
    }
    return true; // continue
  }

  public boolean visit(SingleVariableDeclaration node) {
    IVariableBinding variable = node.resolveBinding();
    if (isNull(variable,
      "null variable in SingleVariableDeclaration \"" + node + "\"")) {
      return true;
    }
    ITypeBinding type = variable.getType();
    if (isNull(type,
      "null type in SingleVariableDeclaration \"" + node + "\"")) {
      return true;
    }
    // declare the variable (it may already have been encountered)
    f_db.declareVariable(variable, type, toJEM(node));
    // if it was added get the variable structure back so we can add to it
    ASTNode parent = node.getParent();
    if (parent != null) {
      if (parent.getNodeType() == ASTNode.METHOD_DECLARATION) {
        // for variables that are a formal parameter of a method or
        // constructor we need to save this information so we can check
        // if any changes are compatible with interface/superclass
        // requirements
        MethodDeclaration method = (MethodDeclaration) parent;
        SingleVariableDeclaration[] parameters =
          (SingleVariableDeclaration[]) method.parameters().toArray(
            new SingleVariableDeclaration[method.parameters().size()]);
        int parameterNumber = -1;
        for (int i = 0; i < parameters.length; i++) {
          if (parameters[i] == node) {
            parameterNumber = i;
          }
        }
        if (parameterNumber == -1) {
          error("parameter not found in method \"" + method + "\"");
          return true;
        }
        IMethodBinding mb = method.resolveBinding();
        if (isNull(mb,
          "null method \"" + method + "\" for actual parameter variable")) {
          return true;
        }
        f_db.addActualParameterInformation(variable, mb, parameterNumber);
      }
    }
    return true;
  }

  //---------------------------------//
  // R E T U R N   S T A T E M E N T //
  //---------------------------------//

  private IMethodBinding currentMethodDeclaration = null;
  private ITypeBinding currentMethodReturnType = null;
  private LinkedList currentMethodDeclarationStack = new LinkedList();
  private LinkedList currentMethodReturnTypeStack = new LinkedList();

  public boolean visit(MethodDeclaration node) {
    currentMethodDeclaration = node.resolveBinding();
    currentMethodReturnType = node.getReturnType().resolveBinding();

    // add to stacks (in parellel)
    currentMethodDeclarationStack.addFirst(currentMethodDeclaration);
    currentMethodReturnTypeStack.addFirst(currentMethodReturnType);
    return true;
  }

  public void endVisit(MethodDeclaration node) {
    if (currentMethodDeclarationStack.size() > 0) {
      currentMethodDeclarationStack.removeFirst();
      currentMethodReturnTypeStack.removeFirst();
    }

    if (currentMethodDeclarationStack.size() > 0) {
      currentMethodDeclaration =
        (IMethodBinding) currentMethodDeclarationStack.getFirst();
      currentMethodReturnType =
        (ITypeBinding) currentMethodReturnTypeStack.getFirst();
    } else {
      currentMethodDeclaration = null;
      currentMethodReturnType = null;
    }
  }

  public boolean visit(ReturnStatement node) {
    Expression expression = node.getExpression();
    if (expression == null) {
      // no expression after "return" -- so we ignore
      return true;
    }
    IVariableBinding[] variable = expressionToVariable(expression);
    if (variable.length == 0) {
      // not interesting, not a variable
      return true;
    }
    for (int i = 0; i < variable.length; i++) {
      ITypeBinding variableType = variable[i].getType();
      if (isNull(variableType,
        "unable to resolve type of " + DOMUtilities.toFullName(variable[i]))) {
        continue;
      }
      if (isNull(currentMethodDeclaration,
        "return statement \"" + node + "\" found outside a method")) {
        continue;
      }
      if (isNull(currentMethodReturnType,
        "return statement \"" + node + "\" found with no known return type")) {
        continue;
      }
      f_db.addToReturnSignature(
        variable[i],
        variableType,
        currentMethodDeclaration,
        currentMethodReturnType,
        toJEM(node));
    }
    return true;
  }

  //---------------------------------------------------------//
  // A S S I G N M E N T   D E P E N D E N C Y   S E A R C H //
  //---------------------------------------------------------//

  public boolean visit(Assignment node) {
    ITypeBinding lhsType = node.getLeftHandSide().resolveTypeBinding();
    if (isNull(lhsType, "null lhs type in Assignment \"" + node + "\"")) {
      return true;
    }
    // ignore primitive types
    if (lhsType.isPrimitive()) {
      return true;
    }
    ITypeBinding rhsType = node.getRightHandSide().resolveTypeBinding();
    if (isNull(rhsType, "null rhs type in Assignment \"" + node + "\"")) {
      return true;
    }
    IVariableBinding rhsVar[] = expressionToVariable(node.getRightHandSide());
    if (rhsVar.length == 0) {
      // some kind of expression we are not interested in
      return true;
    }
    for (int i = 0; i < rhsVar.length; i++) {
      f_db.addToAssignmentSignature(lhsType, rhsVar[i], rhsType, toJEM(node));
    }
    return true;
  }

  public boolean visit(ArrayInitializer node) {
    ITypeBinding arrayType = node.resolveTypeBinding();
    if (isNull(arrayType, "null type for ArrayInitializer \"" + node + "\"")) {
      return false;
    }
    List expList = node.expressions();
    Expression[] expressions =
      (Expression[]) expList.toArray(new Expression[expList.size()]);
    for (int i = 0; i < expressions.length; i++) {
      // evaluate expression for inclusion in type restricton list
      IVariableBinding[] variable = expressionToVariable(expressions[i]);
      if (variable.length > 0) {
        for (int j = 0; j < variable.length; j++) {
          ITypeBinding variableType = variable[j].getType();
          if (isNull(variableType,
            "unable to resolve type of "
              + DOMUtilities.toFullName(variable[j]))) {
            continue;
          }
          // It is a variable we might be interested in
          f_db.addToAssignmentSignature(
            arrayType,
            variable[j],
            variableType,
            toJEM(node));
        }
      }
    }
    return false;
  }

  //-------------------------------------------------------//
  // F I E L D   A N D   M E T H O D   U S E   S E A R C H //
  //-------------------------------------------------------//

  public boolean visit(FieldAccess node) {
    // what is the variable the field access is being invoked on?
    Expression expression = node.getExpression();
    if (isNull(expression,
      "null expression found within FieldAccess \"" + node + "\"")) {
      return true;
    }
    IVariableBinding[] vb = expressionToVariable(expression);
    if (vb.length > 0) {
      for (int i = 0; i < vb.length; i++) {
        ITypeBinding variableType = vb[i].getType();
        if (isNull(variableType,
          "unable to resolve type of " + DOMUtilities.toFullName(vb[i]))) {
          continue;
        }
        IBinding b = node.getName().resolveBinding();
        if (isNull(b, "null field/method binding in \"" + node + "\"")) {
          continue;
        }
        f_db.addToUseSignature(vb[i], variableType, b, toJEM(expression));
      }
    }
    return true;
  }

  public boolean visit(QualifiedName node) {
    // what is the variable the field access is being invoked on?
    Name variableName = node.getQualifier();
    if (isNull(variableName,
      "null qualifier within QualifiedName \"" + node + "\"")) {
      return true;
    }
    IVariableBinding[] vb = expressionToVariable(variableName);
    if (vb.length > 0) {
      for (int i = 0; i < vb.length; i++) {
        ITypeBinding variableType = vb[i].getType();
        if (isNull(variableType,
          "unable to resolve type of " + DOMUtilities.toFullName(vb[i]))) {
          continue;
        }
        IBinding b = node.getName().resolveBinding();
        if (isNull(b, "null field/method binding in \"" + node + "\"")) {
          continue;
        }
        f_db.addToUseSignature(vb[i], variableType, b, toJEM(variableName));
      }
    }
    return true;
  }

  //-----------------------------------------//
  // A C T U A L   P A R A M E T E R   U S E //
  //-----------------------------------------//

  private void examineMethodConstructorInvocationParameters(
    ASTNode node,
    IMethodBinding mb,
    List arguments) {
    Expression[] expressions =
      (Expression[]) arguments.toArray(new Expression[arguments.size()]);
    for (int i = 0; i < expressions.length; i++) {
      // evaluate expression for inclusion in type restricton list
      IVariableBinding[] variable = expressionToVariable(expressions[i]);
      if (variable.length > 0) {
        for (int j = 0; j < variable.length; j++) {
          ITypeBinding variableType = variable[j].getType();
          if (isNull(variableType,
            "unable to resolve type of "
              + DOMUtilities.toFullName(variable[j]))) {
            continue;
          }
          // It is a variable we might be interested in
          f_db.addToActualParameterSignature(
            variable[j],
            variableType,
            mb,
            i,
            toJEM(node));
        }
      }
    }
  }

  public void examineMethodIvocationParameters(MethodInvocation node) {
    IBinding b = node.getName().resolveBinding();
    if (b != null) {
      if (b.getKind() == IBinding.METHOD) {
        IMethodBinding mb = (IMethodBinding) b;
        examineMethodConstructorInvocationParameters(
          node,
          mb,
          node.arguments());
      }
    }
  }

  public boolean visit(MethodInvocation node) {
    // first look at all parameter expressions to see if we care about
    // what is being passed as a parameter to the method
    examineMethodIvocationParameters(node);
    // next look at the variable the method is being invoked on
    // what is the variable the method is being invoked on?
    Expression variableName = node.getExpression();
    if (variableName == null) {
      // null, hopefully always, means we have encountered a local method
      // call (e.g., foo()) and we are not interested in this case
      return true;
    }
    IVariableBinding[] vb = expressionToVariable(variableName);
    if (vb.length > 0) {
      for (int i = 0; i < vb.length; i++) {
        ITypeBinding variableType = vb[i].getType();
        if (isNull(variableType,
          "unable to resolve type of " + DOMUtilities.toFullName(vb[i]))) {
          continue;
        }
        IBinding b = node.getName().resolveBinding();
        if (isNull(b, "null field/method binding in \"" + node + "\"")) {
          continue;
        }
        f_db.addToUseSignature(vb[i], variableType, b, toJEM(variableName));
      }
    }
    return true;
  }

  public boolean visit(SuperMethodInvocation node) {
    IBinding b = node.getName().resolveBinding();
    if (b != null) {
      if (b.getKind() == IBinding.METHOD) {
        IMethodBinding mb = (IMethodBinding) b;
        examineMethodConstructorInvocationParameters(
          node,
          mb,
          node.arguments());
      }
    }
    return true;
  }

  //-------------------------//
  // C O N S T R U C T O R S //
  //-------------------------//

  public boolean visit(ClassInstanceCreation node) {
    IMethodBinding mb = node.resolveConstructorBinding();
    if (isNull(mb,
      "null constructor in ClassInstanceCreation \"" + node + "\"")) {
      return true;
    }
    examineMethodConstructorInvocationParameters(node, mb, node.arguments());
    return true;
  }

  public boolean visit(SuperConstructorInvocation node) {
    IMethodBinding mb = node.resolveConstructorBinding();
    if (isNull(mb,
      "null constructor in SuperConstructorInvocation \"" + node + "\"")) {
      return true;
    }
    examineMethodConstructorInvocationParameters(node, mb, node.arguments());
    return true;
  }

  public boolean visit(ConstructorInvocation node) {
    IMethodBinding mb = node.resolveConstructorBinding();
    if (isNull(mb,
      "null constructor in ConstructorInvocation \"" + node + "\"")) {
      return true;
    }
    examineMethodConstructorInvocationParameters(node, mb, node.arguments());
    return true;
  }

  //  public boolean visit(SimpleName node) {
  //    log("SimpleName: " + node.getIdentifier());
  //    boolean findContext = true;
  //    ASTNode oldParent, parent = node;
  //    while (findContext) {
  //      oldParent = parent;
  //      parent = parent.getParent();
  //      if (parent == null) {
  //        findContext = false;
  //        log(
  //          "SimpleName: FOUND NULL PARENT--last was: "
  //            + oldParent.getClass().getName()
  //            + "\n   "
  //            + oldParent);
  //      } else if (!(parent instanceof Expression)) {
  //        findContext = false;
  //        log(
  //          "SimpleName: FOUND (non expression) PARENT--is: "
  //            + parent.getClass().getName()
  //            + "\n   "
  //            + parent);
  //      }
  //    }
  //    return true;
  //  }

}