package edu.cmu.cs.fluid.Tallyho.analysis.util;

import java.io.PrintStream;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.eclipse.jdt.core.dom.ITypeBinding;

public final class SType {

  // key:String value:SType
  private static Map instanceMap = new WeakHashMap();

  /**
   * dumps instanceMap for debug purposes
   */
  public static void dump(PrintStream out) {
    if (out == null)
      out = System.out; // a reasonable default
    out.println("[SType] B E G I N   I N S T A N C E   D U M P");
    out.println("[SType] Current instance count = " + dumpInstanceCount());
    for (Iterator i = instanceMap.values().iterator(); i.hasNext();) {
      SType e = (SType) i.next();
      out.print("[SType] ");
      if (e.isPrimitive)
        out.print("PRIMITIVE ");
      if (e.isNullType)
        out.print("NULL ");
      if (e.isArray)
        out.print("ARRAY ");
      if (e.isInterface)
        out.print("INTERFACE ");
      else
        out.print("CLASS ");
      if (e.isFromSource)
        out.print("(FROM SRC) ");
      out.print(
        "\""
          + e
          + (e.superclass == null ? "\"" : "\" extends (" + e.superclass + ")"));
      if (e.isArray)
        out.print(
          " dimensions="
            + e.arrayDimensions
            + " element type ("
            + e.arrayElementType
            + ")");
      if (e.interfaces.length > 0) {
        out.print(" implements (");
        for (int j = 0; j < e.interfaces.length; j++) {
          out.print(
            e.interfaces[j] + (j == e.interfaces.length - 1 ? ")" : ","));
        }
      }
      out.println();
      for (int j = 0; j < e.declaredFields.length; j++) {
        out.println("[SType]   F: " + e.declaredFields[j]);
      }
      for (int j = 0; j < e.declaredMethods.length; j++) {
        out.println("[SType]   M: " + e.declaredMethods[j]);
      }
    }
    out.println("[SType] E N D   I N S T A N C E   D U M P");
  }

  public static int dumpInstanceCount() {
    return instanceMap.size();
  }

  public static SType getInstance(ITypeBinding t) {
    if (t == null) {
      throw new NullPointerException();
    }
    SType newInstance = new SType();
    // name is the instance key
    newInstance.name = DOMUtilities.toFullName(t);
    if (instanceMap.containsKey(newInstance.name)) {
      return (SType) instanceMap.get(newInstance.name);
    } else {
      instanceMap.put(newInstance.name, newInstance);
      List interfaces = new LinkedList();
      for (int i = 0; i < t.getInterfaces().length; i++) {
        SType aInterface = SType.getInstance(t.getInterfaces()[i]);
        if (aInterface != newInstance) {
          // avoid nasty bug where an interfaces is it's own interface
          interfaces.add(aInterface);
        } else {
          System.out.println(
            "[SType] "
              + newInstance.name
              + " thinks it is an interface of itself");
        }
      }
      newInstance.interfaces =
        (SType[]) interfaces.toArray(new SType[interfaces.size()]);
      newInstance.isInterface = t.isInterface();
      newInstance.declaredFields = new SVariable[t.getDeclaredFields().length];
      for (int i = 0; i < t.getDeclaredFields().length; i++) {
        newInstance.declaredFields[i] =
          SVariable.getInstance(t.getDeclaredFields()[i]);
      }
      newInstance.declaredMethods = new SMethod[t.getDeclaredMethods().length];
      for (int i = 0; i < t.getDeclaredMethods().length; i++) {
        newInstance.declaredMethods[i] =
          SMethod.getInstance(t.getDeclaredMethods()[i]);
      }
      ITypeBinding superclass = t.getSuperclass();
      newInstance.superclass =
        (superclass == null ? null : SType.getInstance(superclass));
      newInstance.isInterface = t.isInterface();
      newInstance.isArray = t.isArray();
      newInstance.arrayDimensions = t.getDimensions();
      ITypeBinding elementType = t.getElementType();
      newInstance.arrayElementType =
        (elementType == null ? null : SType.getInstance(elementType));
      newInstance.isPrimitive = t.isPrimitive();
      newInstance.isNullType = t.isNullType();
      newInstance.isFromSource = t.isFromSource();
      // quick check for nulls
      if (newInstance.anyNull()) {
        System.out.println("[SType] NULL FIELDS DETECTED");
      }
      return newInstance;
    }
  }

  private int arrayDimensions = 0;
  private SType arrayElementType;
  private SVariable[] declaredFields = new SVariable[0];
  private SMethod[] declaredMethods = new SMethod[0];
  private SType[] interfaces = new SType[0];
  private boolean isArray;
  private boolean isFromSource;
  private boolean isInterface;
  private boolean isNullType;
  private boolean isPrimitive;
  private String name = "";
  private SType superclass;

  private SType() {
  }

  private boolean anyNull() {
    if (declaredFields == null || declaredMethods == null || name == null) {
      return true;
    }
    if (isArray && arrayElementType == null) {
      return true;
    }
    if (!isInterface
      && !isPrimitive
      && !isNullType
      && !isArray
      && superclass == null
      && name.intern() != "java.lang.Object") {
      return true;
    }
    return false;
  }

  public boolean containsField(SVariable f) {
    for (int i = 0; i < declaredFields.length; i++) {
      if (f == declaredFields[i]) {
        return true;
      }
    }
    if (superclass != null) {
      if (superclass.containsField(f)) {
        return true;
      }
    }
    return false;
  }

  public boolean containsMethod(SMethod m) {
    if (m == null) {
      return false;
    }
    for (int i = 0; i < declaredMethods.length; i++) {
      if (m.signatureMatches(declaredMethods[i])) {
        return true;
      }
    }
    if (superclass != null) {
      if (superclass.containsMethod(m)) {
        return true;
      }
    }
    for (int i = 0; i < interfaces.length; i++) {
      if (interfaces[i].containsMethod(m)) {
        return true;
      }
    }
    return false;
  }

  public SVariable[] getDeclaredFields() {
    return declaredFields;
  }

  public SMethod[] getDeclaredMethods() {
    return declaredMethods;
  }

  public int getDimensions() {
    return arrayDimensions;
  }

  public SType getElementType() {
    return arrayElementType;
  }

  public SType[] getInterfaces() {
    return interfaces;
  }

  public String getName() {
    return name;
  }

  public SType getSuperclass() {
    return superclass;
  }

  public boolean isArray() {
    return isArray;
  }

  public boolean isFromSource() {
    return isFromSource;
  }

  public boolean isInterface() {
    return isInterface;
  }

  public boolean isNullType() {
    return isNullType;
  }

  public boolean isPrimitive() {
    return isPrimitive;
  }

  public boolean isSubtypeOf(String typeName) {
    if (name.equals(typeName)) {
      return true;
    }
    if (superclass != null) {
      if (superclass.isSubtypeOf(typeName)) {
        return true;
      }
    }
    for (int i = 0; i < interfaces.length; i++) {
      if (interfaces[i].isSubtypeOf(typeName)) {
        return true;
      }
    }
    return false;
  }

  public boolean isSubtypeOf(SType t) {
    if (this == t) {
      return true;
    }
    if (superclass != null) {
      if (superclass.isSubtypeOf(t)) {
        return true;
      }
    }
    for (int i = 0; i < interfaces.length; i++) {
      if (interfaces[i].isSubtypeOf(t)) {
        return true;
      }
    }
    return false;
  }

  public String toString() {
    return name;
  }
}